from __future__ import annotations

from typing import TYPE_CHECKING
from typing import TypeVar

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any

    import svg
import numpy as np

from musiclib import config
from musiclib.interval import AbstractInterval
from musiclib.svg.reprsvg import ReprSVGMixin
from musiclib.util.cache import Cached

Self = TypeVar('Self', bound='IntervalSet')


class IntervalSet(Cached, ReprSVGMixin):
    def __init__(self, intervals: frozenset[AbstractInterval]) -> None:
        if not isinstance(intervals, frozenset):
            raise TypeError(f'expected frozenset, got {type(intervals)}')
        if any(not isinstance(interval, AbstractInterval) for interval in intervals):
            raise TypeError('expected AbstractInterval items')
        self.intervals = intervals
        self.intervals_ascending = tuple(sorted(intervals))
        self.bits = ''.join('1' if AbstractInterval(i) in intervals else '0' for i in range(12))
        self.names: frozenset[str] = config.intervals_to_names.get(intervals, frozenset())
        self.name_kinds = {name: config.kinds[name] for name in self.names}

    @classmethod
    def from_name(cls: type[Self], name: str) -> Self:
        return cls(config.name_to_intervals[name])

    @classmethod
    def from_bits(cls, bits: str) -> IntervalSet:
        return cls(frozenset(AbstractInterval(i) for i, v in enumerate(bits) if int(v)))

    @classmethod
    def from_base12(cls, intervals: frozenset[str]) -> IntervalSet:
        return cls(frozenset(AbstractInterval.from_str(i) for i in intervals))

    @property
    def inverse(self) -> IntervalSet:
        return IntervalSet(frozenset(-i for i in self.intervals))

    def __len__(self) -> int:
        return len(self.intervals)

    def __iter__(self) -> Iterator[AbstractInterval]:
        return iter(self.intervals_ascending)

    def __getnewargs__(self) -> tuple[frozenset[AbstractInterval]]:
        return (self.intervals,)

    def __str__(self) -> str:
        return '_'.join(str(i) for i in self.intervals_ascending)

    def __repr__(self) -> str:
        return f"IntervalSet({' '.join(np.base_repr(i.interval, base=12) for i in self.intervals_ascending)})"

    def svg_piano(self, **kwargs: Any) -> svg.SVG:
        raise NotImplementedError('IntervalSet does not have a piano representation')

    def svg_plane_piano(self, **kwargs: Any) -> svg.SVG:
        from musiclib.svg.card import PlanePiano
        kwargs.setdefault('interval_colors', {i: config.interval_colors[i] for i in self.intervals})
        kwargs.setdefault('header_kwargs', {'title': str(self)})
        return PlanePiano(**kwargs).svg



from __future__ import annotations

import itertools
import random
from collections.abc import Sequence
from typing import TYPE_CHECKING
from typing import Any
from typing import TypeVar
from typing import overload

from musiclib import config
from musiclib.note import Note
from musiclib.note import SpecificNote
from musiclib.svg.isomorphic.text import FromIntervalDict
from musiclib.svg.reprsvg import ReprSVGMixin
from musiclib.util.cache import Cached

if TYPE_CHECKING:
    from collections.abc import Iterator

    import svg


Self = TypeVar('Self', bound='NoteSet')


class NoteSet(Cached, ReprSVGMixin):
    def __init__(self, notes: frozenset[Note]) -> None:
        if not isinstance(notes, frozenset):
            raise TypeError(f'expected frozenset, got {type(notes)}')
        self.notes = notes
        self.notes_ascending = tuple(sorted(self.notes))
        self.note_to_intervals = {left: frozenset(right - left for right in self.notes) for left in self.notes}
        self.intervals_key = frozenset(self.note_to_intervals.values())
        self.name = config.intervals_key_to_name.get(self.intervals_key)
        self._note_i = {note: i for i, note in enumerate(self.notes_ascending)}

    @classmethod
    def from_str(cls: type[Self], string: str) -> Self:
        return cls(frozenset(Note(note) for note in string))

    @overload
    def add_note(self, note: SpecificNote, steps: int) -> SpecificNote:
        ...

    @overload
    def add_note(self, note: Note, steps: int) -> Note:
        ...

    def add_note(self, note: Note | SpecificNote, steps: int) -> Note | SpecificNote:
        if len(self) == 0:
            raise NotImplementedError('cannot add notes using empty noteset')
        notes = self.notes_ascending
        if isinstance(note, Note):
            return notes[(notes.index(note) + steps) % len(notes)]
        if isinstance(note, SpecificNote):
            octaves, i = divmod(self.notes_ascending.index(note.abstract) + steps, len(self.notes))
            return SpecificNote(self.notes_ascending[i], octave=note.octave + octaves)
        raise TypeError

    def subtract(self, left: Note | SpecificNote, right: Note | SpecificNote) -> int:
        if type(left) is type(right) is Note:
            return (self._note_i[left] - self._note_i[right]) % len(self)
        if type(left) is type(right) is SpecificNote:
            return self._note_i[left.abstract] - self._note_i[right.abstract] + len(self) * (left.octave - right.octave)
        raise TypeError('left and right should be either Note or SpecificNote')

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NoteSet):
            raise TypeError
        return self.notes == other.notes

    def __hash__(self) -> int:
        return hash(self.notes)

    def __len__(self) -> int:
        return len(self.notes)

    def __getitem__(self, item: int) -> Note:
        return self.notes_ascending[item]

    def __iter__(self) -> Iterator[Note]:
        return iter(self.notes_ascending)

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, Note):
            raise TypeError
        return item in self.notes

    def __le__(self, other: object) -> bool:
        if not isinstance(other, NoteSet):
            raise TypeError
        return self.notes <= other.notes

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, NoteSet):
            raise TypeError
        return other.notes <= self.notes

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, NoteSet):
            raise TypeError
        return self.notes < other.notes

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, NoteSet):
            raise TypeError
        return other.notes < self.notes

    def __str__(self) -> str:
        return ''.join(note.name for note in self)

    def __repr__(self) -> str:
        return f"NoteSet('{self}')"

    def svg_piano(self, **kwargs: Any) -> svg.SVG:
        from musiclib.svg.card import Piano  # hack to fix circular import
        kwargs.setdefault('header_kwargs', {'title': str(self)})
        kwargs.setdefault(
            'regular_piano_kwargs', {
                'note_colors': {note: config.RED for note in self},
            },
        )
        return Piano(**kwargs).svg

    def svg_plane_piano(self, **kwargs: Any) -> svg.SVG:
        from musiclib.svg.card import PlanePiano
        if self.notes:
            kwargs.setdefault(
                'interval_colors', {
                    i: config.RED
                    for i in self.note_to_intervals[self.notes_ascending[0]]
                },
            )
            kwargs.setdefault(
                'interval_text', FromIntervalDict(
                    {
                        note - self.notes_ascending[0]: str(note)
                        for note in self.notes_ascending
                    }, abstract=True,
                ),
            )
        kwargs.setdefault('header_kwargs', {'title': str(self)})
        return PlanePiano(**kwargs).svg

    def __getnewargs__(self) -> tuple[frozenset[Note]]:
        return (self.notes,)


CHROMATIC_NOTESET = NoteSet.from_str(config.chromatic_notes)


class SpecificNoteSet(Cached, ReprSVGMixin, Sequence[SpecificNote]):
    def __init__(self, notes: frozenset[SpecificNote]) -> None:
        if not isinstance(notes, frozenset):
            raise TypeError(f'expected frozenset, got {type(notes)}')
        self.notes = notes
        self.noteset = NoteSet(frozenset(note.abstract for note in notes))
        self.notes_ascending = tuple(sorted(notes))
        self.intervals = tuple(note - self.notes_ascending[0] for note in self.notes_ascending)  # from lowest note

    @classmethod
    def random(
        cls,
        n_notes: int | None = None,
        notes: frozenset[Note] = frozenset({Note(n) for n in config.chromatic_notes}),  # noqa: B008
        octaves: frozenset[int] = frozenset({3, 4, 5}),
    ) -> SpecificNoteSet:
        if n_notes is None:
            n_notes = random.randint(2, 5)
        notes_space = tuple(
            SpecificNote(note, octave)
            for note, octave in itertools.product(notes, octaves)
        )
        return cls(frozenset(random.sample(notes_space, n_notes)))

    @classmethod
    def from_str(cls, string: str) -> SpecificNoteSet:
        notes_ = string.split('_')
        if len(notes_) != len(set(notes_)):
            raise ValueError('SpecificNoteSet with non unique notes are not supported')
        notes = frozenset(SpecificNote.from_str(note) for note in notes_)
        return cls(notes)

    @classmethod
    def from_noterange(
        cls,
        start: SpecificNote,
        stop: SpecificNote,
        noteset: NoteSet = CHROMATIC_NOTESET,
    ) -> SpecificNoteSet:
        if not (isinstance(start, SpecificNote) and isinstance(stop, SpecificNote)):
            raise TypeError('start and stop should be SpecificNote instances')
        if start > stop:  # both ends included
            raise ValueError('start should be <= stop')
        if not {start.abstract, stop.abstract} <= noteset.notes:
            raise KeyError('start and stop notes should be in the noteset')
        return cls(
            frozenset(
                noteset.add_note(start, i)
                for i in range(noteset.subtract(stop, start) + 1)
            ),
        )
        # return cls(frozenset(note for note in NoteRange(start, stop) if note.abstract in noteset)))

    def notes_combinations(self) -> Iterator[tuple[SpecificNote, SpecificNote]]:
        yield from itertools.combinations(self.notes_ascending, 2)

    def find_intervals(self, interval: int) -> tuple[tuple[SpecificNote, SpecificNote], ...]:
        return tuple((n, m) for n, m in self.notes_combinations() if abs(m - n) == interval)

    def transpose_to_note(self, note: SpecificNote) -> SpecificNoteSet:
        if len(self) == 0:
            return self
        return self + (note - self[0])

    def __len__(self) -> int:
        return len(self.notes)

    @overload
    def __getitem__(self, i: int) -> SpecificNote:
        ...

    @overload
    def __getitem__(self, s: slice) -> SpecificNoteSet:
        ...

    # TODO: try here SpecificNote | SpecificNoteSet
    def __getitem__(self, item: int | slice) -> SpecificNote | Sequence[SpecificNote]:
        if isinstance(item, slice):
            return SpecificNoteSet(frozenset(self.notes_ascending[item]))
        return self.notes_ascending[item]

    def __iter__(self) -> Iterator[SpecificNote]:
        return iter(self.notes_ascending)

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, SpecificNote):
            raise TypeError
        return item in self.notes

    def __str__(self) -> str:
        return '_'.join(str(note) for note in self.notes_ascending)

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SpecificNoteSet):
            raise TypeError
        return self.notes == other.notes

    def __hash__(self) -> int:
        return hash(self.notes)

    def __sub__(self, other: SpecificNoteSet) -> int:
        return sum(abs(a - b) for a, b in zip(self, other, strict=True))

    def __add__(self, other: int) -> SpecificNoteSet:
        """transpose"""
        if not isinstance(other, int):
            raise TypeError('only adding integers is allowed (transposition)')
        return SpecificNoteSet(frozenset(note + other for note in self))

    def __getnewargs__(self) -> tuple[frozenset[SpecificNote]]:
        return (self.notes,)

    def svg_piano(self, **kwargs: Any) -> svg.SVG:
        from musiclib.svg.card import Piano
        kwargs.setdefault('header_kwargs', {'title': str(self)})
        kwargs.setdefault(
            'regular_piano_kwargs', {
                'note_colors': dict.fromkeys(self.notes, config.RED),
                'squares': {note: {'text': str(note), 'text_size': '8'} for note in self},
                'start_stop': (self[0], self[-1]) if self.notes else None,
            },
        )
        return Piano(**kwargs).svg

    def svg_plane_piano(self, **kwargs: Any) -> svg.SVG:
        from musiclib.svg.card import PlanePiano
        if self.notes:
            kwargs.setdefault(
                'interval_colors', {
                    i: config.RED
                    for i in self.intervals
                },
            )
            kwargs.setdefault(
                'interval_text', FromIntervalDict({
                    interval: str(note)
                    for interval, note in zip(self.intervals, self.notes_ascending, strict=True)
                }),
            )
            kwargs.setdefault('n_cols', max(self) - min(self) + 1)
        kwargs.setdefault('header_kwargs', {'title': str(self)})
        return PlanePiano(**kwargs).svg


def subsets(noteset: NoteSet, min_notes: int = 1) -> frozenset[NoteSet]:
    out = set()
    for n_subset in range(min_notes, len(noteset) + 1):
        for notes in itertools.combinations(noteset, n_subset):
            out.add(NoteSet(frozenset(notes)))
    return frozenset(out)


class ComparedNoteSets(Cached, ReprSVGMixin):
    """
    this is compared scale
    local terminology: left scale is compared to right
    left is kinda parent, right is kinda child
    """

    def __init__(self, left: NoteSet, right: NoteSet) -> None:
        if not (isinstance(left, NoteSet) and isinstance(right, NoteSet)):
            raise TypeError(f'expected NoteSet, got {type(left)} and {type(right)}')
        self.left = left
        self.right = right
        self.key = left, right
        self.shared_notes = frozenset(left.notes) & frozenset(right.notes)
        self.new_notes = frozenset(right.notes) - frozenset(left.notes)
        self.del_notes = frozenset(left.notes) - frozenset(right.notes)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ComparedNoteSets):
            raise TypeError
        return self.key == other.key

    def __hash__(self) -> int:
        return hash(self.key)

    def __str__(self) -> str:
        return f'{self.left} | {self.right}'

    def __repr__(self) -> str:
        return f'ComparedNoteSets({self})'

    def svg_piano(self, **kwargs: Any) -> svg.SVG:
        from musiclib.svg.card import Piano
        kwargs.setdefault(
            'regular_piano_kwargs', {
                'note_colors':
                dict.fromkeys(self.del_notes, config.RED) |
                dict.fromkeys(self.new_notes, config.GREEN) |
                dict.fromkeys(self.shared_notes, config.BLUE),
            },
        )

        kwargs.setdefault('header_kwargs', {'title': str(self)})
        return Piano(**kwargs).svg

    def svg_plane_piano(self, **kwargs: Any) -> svg.SVG:
        from musiclib.svg.card import PlanePiano

        n0 = Note(config.chromatic_notes[0])
        kwargs.setdefault('header_kwargs', {'title': str(self)})
        kwargs.setdefault(
            'interval_colors',
            dict.fromkeys([n - n0 for n in self.del_notes], config.RED) |
            dict.fromkeys([n - n0 for n in self.new_notes], config.GREEN) |
            dict.fromkeys([n - n0 for n in self.shared_notes], config.BLUE),
        )
        kwargs.setdefault('interval_text', FromIntervalDict({n - n0: str(n) for n in CHROMATIC_NOTESET}, abstract=True))
        return PlanePiano(**kwargs).svg



from typing import Literal

import svg
from colortool import Color

from musiclib import config


class Header:
    def __init__(
        self,
        *,
        title: str | None = None,
        subtitle: str | None = None,
        title_href: str | None = None,
        title_font_size: int = 15,
        subtitle_href: str | None = None,
        subtitle_font_size: int = 12,
        title_y: int = 4,
        subtitle_y: int = 18,
        background_color: Color = config.WHITE_BRIGHT,
        width: int = 300,
        height: int = 35,
        margin: tuple[int, int, int, int] = (3, 3, 3, 3),
        padding: tuple[int, int, int, int] = (30, 2, 2, 2),
        border_radius: int = 0,
        header_rect: bool = True,
    ) -> None:
        self.width = width
        self.height = height
        self.elements: list[svg.Element] = []
        if header_rect:
            self.elements.append(
                svg.Rect(
                    class_=['header_rect'],
                    x=0,
                    y=0,
                    width=width,
                    height=height,
                    fill=background_color.css_hex,
                    rx=border_radius,
                    ry=border_radius,
                    stroke_width=1,
                    stroke=config.BLACK_PALE.css_hex,
                ),
            )

        if title is not None:
            dominant_baseline: Literal['hanging', 'central']
            if subtitle is not None:
                y = margin[0] + title_y
                dominant_baseline = 'hanging'
            else:
                y = (margin[0] + padding[0]) // 2
                dominant_baseline = 'central'
            text_title = svg.Text(
                x=margin[3] + padding[3],
                y=y,
                text=title,
                font_size=title_font_size,
                font_weight='bold',
                dominant_baseline=dominant_baseline,
            )
            if title_href is not None:
                self.elements.append(svg.A(href=title_href, elements=[text_title]))
            else:
                self.elements.append(text_title)
        if subtitle is not None:
            text_subtitle = svg.Text(
                x=margin[3] + padding[3],
                y=margin[0] + subtitle_y,
                font_size=subtitle_font_size,
                fill=config.BLACK_BRIGHT.css_hex,
                text=subtitle,
                dominant_baseline='hanging',
            )
            if subtitle_href is not None:
                self.elements.append(svg.A(href=subtitle_href, elements=[text_subtitle]))
            else:
                self.elements.append(text_subtitle)

    @property
    def svg(self) -> svg.SVG:
        return svg.SVG(width=self.width, height=self.height, elements=self.elements)

    def _repr_svg_(self) -> str:
        return str(self.svg)



from __future__ import annotations

import typing as tp

import mido

DEFAULT_TICKS_PER_BEAT = 96
DEFAULT_BEATS_PER_BAR = 4
DEFAULT_BEATS_PER_MINUTE = 120


class Tempo:
    def __init__(
        self,
        ticks: int = 0,
        ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT,
        beats_per_bar: int = DEFAULT_BEATS_PER_BAR,
        beats_per_minute: float = DEFAULT_BEATS_PER_MINUTE,
    ) -> None:
        self.ticks = ticks
        self.ticks_per_beat = ticks_per_beat
        self.beats_per_bar = beats_per_bar
        self.beats_per_minute = beats_per_minute

    @classmethod
    def from_beats(
        cls,
        beats: float,
        ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT,
        **kwargs: tp.Any,
    ) -> Tempo:
        return cls(
            ticks=int(beats * ticks_per_beat),
            ticks_per_beat=ticks_per_beat,
            **kwargs,
        )

    @classmethod
    def from_bars(
        cls,
        bars: float,
        ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT,
        beats_per_bar: int = DEFAULT_BEATS_PER_BAR,
        **kwargs: tp.Any,
    ) -> Tempo:
        return cls(
            ticks=int(bars * beats_per_bar * ticks_per_beat),
            ticks_per_beat=ticks_per_beat,
            beats_per_bar=beats_per_bar,
            **kwargs,
        )

    @classmethod
    def from_seconds(
        cls,
        seconds: float,
        ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT,
        beats_per_minute: float = DEFAULT_BEATS_PER_MINUTE,
        **kwargs: tp.Any,
    ) -> Tempo:
        return cls(
            ticks=int(seconds * beats_per_minute / 60 * ticks_per_beat),
            ticks_per_beat=ticks_per_beat,
            beats_per_minute=beats_per_minute,
            **kwargs,
        )

    @property
    def beats(self) -> float:
        return self.ticks / self.ticks_per_beat

    @property
    def bars(self) -> float:
        return self.beats / self.beats_per_bar

    @property
    def seconds(self) -> float:
        return self.beats / self.beats_per_minute * 60

    @property
    def beats_per_second(self) -> float:
        return self.beats_per_minute / 60

    @property
    def ticks_per_second(self) -> float:
        return self.ticks_per_beat * self.beats_per_second

    @property
    def bars_per_second(self) -> float:
        return self.beats_per_second / self.beats_per_bar

    @property
    def ticks_per_bar(self) -> int:
        return self.ticks_per_beat * self.beats_per_bar

    @property
    def midi_tempo(self) -> int:
        return mido.bpm2tempo(bpm=self.beats_per_minute, time_signature=(self.beats_per_bar, 4))  # type: ignore[no-any-return]

    def __repr__(self) -> str:
        return f'Tempo(ticks={self.ticks!r}, ticks_per_beat={self.ticks_per_beat!r}, beats_per_bar={self.beats_per_bar!r}, beats_per_minute={self.beats_per_minute!r})'



import abc
import uuid
from collections.abc import Iterable
from typing import Literal

import svg
from colortool import Color

from musiclib import config
from musiclib.interval import AbstractInterval
from musiclib.svg.isomorphic.text import TEXT_CALLABLE
from musiclib.svg.isomorphic.text import middle_text_kw_abstract_interval
from musiclib.util.etc import are_mutually_exclusive
from musiclib.util.etc import is_any_not_none
from musiclib.util.etc import vertex


class IsomorphicKeyboard(abc.ABC):
    """
    https://notes.tandav.me/notes/3548
    """

    def __init__(
        self,
        *,
        interval_colors: dict[AbstractInterval | int, Color] | None = None,
        interval_strokes: dict[AbstractInterval | int, Color] | None = None,
        interval_radial_parts_colors: dict[int, dict[int, Color]] | None = None,
        interval_horizontal_parts_colors: dict[int, dict[int, Color]] | None = None,
        interval_vertical_parts_colors: dict[int, dict[int, Color]] | None = None,
        n_parts: int | None = None,
        interval_text: TEXT_CALLABLE | None = middle_text_kw_abstract_interval,
        interval_subtext: TEXT_CALLABLE | None = None,
        interval_extra_texts: Iterable[TEXT_CALLABLE] = (),
        n_rows: int | None = 7,
        n_cols: int | None = 13,
        row_range: range | None = None,
        col_range: range | None = None,
        ax0_step: int = 2,
        ax1_step: int = 1,
        radius: int = 18,
        offset_x: int = 0,
        offset_y: int = 0,
        round_points: bool = True,
        rotated: bool = False,
        default_key_color: Color = config.BLACK_PALE,
    ) -> None:
        self.rotated = rotated
        self.validate_dimensions(n_rows, n_cols, row_range, col_range)
        if self.n_rows == 0 or self.n_cols == 0:
            return
        self.radius = radius
        self.offset_x = offset_x
        self.offset_y = offset_y
        self.elements: list[svg.Element] = []
        self.interval_colors = interval_colors or {}

        self.validate_parts_colors(
            interval_radial_parts_colors,
            interval_horizontal_parts_colors,
            interval_vertical_parts_colors,
            n_parts,
        )
        self.interval_text = interval_text
        self.interval_subtext = interval_subtext
        self.interval_extra_texts = interval_extra_texts
        self.round_points = round_points
        self.interval_strokes = interval_strokes or {}
        self.defs = svg.Defs(elements=[])
        self.elements.append(self.defs)
        self.id_suffix = str(uuid.uuid4()).split('-', maxsplit=1)[0]
        self.ax0_step = ax0_step
        self.ax1_step = ax1_step
        self.default_key_color = default_key_color
        self.add_keys()

    def validate_parts_colors(
        self,
        interval_radial_parts_colors: dict[int, dict[int, Color]] | None = None,
        interval_horizontal_parts_colors: dict[int, dict[int, Color]] | None = None,
        interval_vertical_parts_colors: dict[int, dict[int, Color]] | None = None,
        n_parts: int | None = None,
    ) -> None:
        if not are_mutually_exclusive(
            interval_radial_parts_colors,
            interval_horizontal_parts_colors,
            interval_vertical_parts_colors,
        ):
            raise ValueError('Exactly one of interval_radial_parts_colors, interval_horizontal_parts_colors, interval_vertical_parts_colors must be provided')

        if is_any_not_none(
            interval_radial_parts_colors,
            interval_horizontal_parts_colors,
            interval_vertical_parts_colors,
        ) and n_parts is None:
            raise ValueError('n_parts must be provided if any of interval_radial_parts_colors, interval_horizontal_parts_colors, interval_vertical_parts_colors is provided')

        self.interval_radial_parts_colors = interval_radial_parts_colors
        self.interval_horizontal_parts_colors = interval_horizontal_parts_colors
        self.interval_vertical_parts_colors = interval_vertical_parts_colors
        self.n_parts = n_parts

    def validate_dimensions(
        self,
        n_rows: int | None,
        n_cols: int | None,
        row_range: range | None,
        col_range: range | None,
    ) -> None:
        if not are_mutually_exclusive(n_rows, row_range):
            raise ValueError('Exactly one of n_rows or row_range must be provided')
        if not are_mutually_exclusive(n_cols, col_range):
            raise ValueError('Exactly one of n_cols or col_range must be provided')

        self.row_range = row_range
        self.col_range = col_range

        if n_rows is None:
            if row_range is None:
                raise ValueError('Exactly one of n_rows or row_range must be provided')
            self.n_rows = len(row_range)
        elif n_rows < 0:
            raise ValueError(f'n_rows={n_rows} must be positive')
        else:
            self.n_rows = n_rows

        if n_cols is None:
            if col_range is None:
                raise ValueError('Exactly one of n_cols or col_range must be provided')
            self.n_cols = len(col_range)
        elif n_cols < 0:
            raise ValueError(f'n_cols={n_cols} must be positive')
        else:
            self.n_cols = n_cols

    @abc.abstractmethod
    def add_keys(self) -> None:
        ...

    @abc.abstractmethod
    def col_to_x(self, col: float) -> float:
        ...

    @abc.abstractmethod
    def row_to_y(self, row: float) -> float:
        ...

    @property
    @abc.abstractmethod
    def width(self) -> int:
        ...

    @property
    @abc.abstractmethod
    def height(self) -> int:
        ...

    @abc.abstractmethod
    def row_col_to_interval(self, row: float, col: float) -> int:
        ...

    def add_parts(self, interval: int, x: float, y: float, id_: str) -> None:  # noqa: C901
        if self.n_parts is None:
            raise ValueError('n_parts must be provided if any of interval_radial_parts_colors, interval_horizontal_parts_colors, interval_vertical_parts_colors is provided')
        if self.interval_radial_parts_colors is not None:
            for part, color in self.interval_radial_parts_colors.get(interval, {}).items():
                if part >= self.n_parts:
                    raise ValueError(f'part={part} must be less than n_parts={self.n_parts}')
                p0 = vertex(x, y, self.radius, self.n_parts, part)
                p1 = vertex(x, y, self.radius, self.n_parts, (part + 1) % self.n_parts)
                self.elements.append(
                    svg.Path(
                        d=[
                            svg.MoveTo(x, y),
                            svg.LineTo(*p0),
                            svg.Arc(rx=self.radius, ry=self.radius, angle=360 / self.n_parts, large_arc=False, sweep=False, x=p1[0], y=p1[1]),
                            svg.ClosePath(),
                        ],
                        fill=color.css_hex,
                        clip_path=f'url(#{id_})',
                    ),
                )

        if self.interval_horizontal_parts_colors is not None:
            for part, color in self.interval_horizontal_parts_colors.get(interval, {}).items():
                if part >= self.n_parts:
                    raise ValueError(f'part={part} must be less than n_parts={self.n_parts}')
                self.elements.append(
                    svg.Rect(
                        **self.ax_split_part_rect_coordinates(x, y, part, 'horizontal'),  # type: ignore[arg-type]
                        fill=color.css_hex,
                        clip_path=f'url(#{id_})',
                    ),
                )

        if self.interval_vertical_parts_colors is not None:
            for part, color in self.interval_vertical_parts_colors.get(interval, {}).items():
                if part >= self.n_parts:
                    raise ValueError(f'part={part} must be less than n_parts={self.n_parts}')
                self.elements.append(
                    svg.Rect(
                        **self.ax_split_part_rect_coordinates(x, y, part, 'vertical'),  # type: ignore[arg-type]
                        fill=color.css_hex,
                        clip_path=f'url(#{id_})',
                    ),
                )

    @abc.abstractmethod
    def ax_split_part_rect_coordinates(self, x: float, y: float, part: int, ax: Literal['horizontal', 'vertical']) -> dict[str, float]:
        ...

    def add_key(self, row: float, col: float) -> None:
        interval = self.row_col_to_interval(row, col)
        x = self.col_to_x(col)
        y = self.row_to_y(row)
        color = self.interval_colors.get(interval, self.interval_colors.get(AbstractInterval(interval), self.default_key_color))
        points = self.key_points(x, y)
        if self.round_points:
            points = [round(p, 1) for p in points]

        polygon_kw = {
            'class_': ['polygon-colored'],
            'fill': color.css_hex,
            'points': points,
            # clip_path='url(#key-clip)',
        }
        id_ = f'row-{row}-col-{col}-{self.id_suffix}'
        self.defs.elements.append(svg.ClipPath(id=id_, elements=[svg.Polygon(**polygon_kw)]))  # type: ignore[union-attr]
        stroke = self.interval_strokes.get(interval, self.interval_strokes.get(AbstractInterval(interval), None))
        if stroke is not None:
            polygon_kw['stroke'] = stroke['stroke'].css_hex
            polygon_kw['stroke_width'] = stroke.get('stroke_width', 1) * 2
            polygon_kw['clip_path'] = f'url(#{id_})'
        self.elements.append(svg.Polygon(**polygon_kw))

        if self.n_parts is not None:
            self.add_parts(interval, x, y, id_)

        for text_callable in (
            self.interval_text,
            self.interval_subtext,
            *self.interval_extra_texts,
        ):
            if text_callable is None:
                continue
            value = text_callable(interval, self.radius, x, y)
            if value is None:
                continue
            self.elements.append(svg.Text(**value))

        # transparent polygon on top for mouse events
        # polygon = svg.Polygon(
        #     class_=['polygon-transparent'],
        #     points=points,
        #     fill=Color.from_rgba_int((0, 0, 0, 0)).css_rgba,
        #     # stroke='black',
        #     # stroke_width=1,
        #     # onmousedown=f"midi_message('note_on', '{note}')",
        #     # onmouseup=f"midi_message('note_off', '{note}')",
        # )
        # self.elements.append(polygon)

    @abc.abstractmethod
    def key_points(self, x: float, y: float) -> list[float]:
        ...

    @property
    def svg(self) -> svg.SVG:
        return svg.SVG(
            width=self.width,
            height=self.height,
            elements=self.elements,
        )

    def _repr_svg_(self) -> str:
        return str(self.svg)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import TypedDict

import svg

from musiclib import config
from musiclib.note import Note
from musiclib.note import SpecificNote
from musiclib.noteset import SpecificNoteSet

if TYPE_CHECKING:
    from colortool import Color


def note_color(note: Note | SpecificNote) -> Color:
    def _note_color(note: Note) -> Color:
        return config.BLACK_PALE if note.is_black else config.WHITE_PALE
    if isinstance(note, SpecificNote):
        return _note_color(note.abstract)
    if isinstance(note, Note):
        return _note_color(note)
    raise TypeError


class SquaresPayload(TypedDict, total=False):
    fill_color: Color
    border_color: Color
    text_color: Color
    text_size: Color
    text: str
    onclick: str
    font_family: str


class RegularPiano:
    def __init__(
        self,
        *,
        note_colors: dict[Note | SpecificNote, Color] | None = None,
        note_hrefs: dict[Note | SpecificNote, Color] | None = None,
        note_onclicks: dict[Note | SpecificNote, Color] | None = None,
        top_rect_colors: dict[Note | SpecificNote, Color] | None = None,
        squares: dict[Note | SpecificNote, SquaresPayload] | None = None,
        start_stop: tuple[SpecificNote, SpecificNote] | None = None,
        ww: int = 18,  # white key width
        wh: int = 85,  # white key height
        black_small_width_ratio: float = 0.6,
        black_small_height_ratio: float = 0.6,
        top_rect_height: int = 5,
        square_size: int = 12,
        square_white_offset: int = 5,
        square_black_offset: int = 3,
        class_: tuple[str, ...] = (),
        id: str | None = None,  # noqa: A002 # pylint: disable=redefined-builtin
    ) -> None:
        self.note_colors = note_colors or {}
        self.note_hrefs = note_hrefs or {}
        self.note_onclicks = note_onclicks or {}
        self.top_rect_colors: dict[Note | SpecificNote, Color] = top_rect_colors or {}
        self.squares = squares or {}
        self.ww = ww
        self.wh = wh
        self.bw = int(ww * black_small_width_ratio)
        self.bh = int(wh * black_small_height_ratio)
        self.top_rect_height = top_rect_height
        self.square_size = square_size
        self.square_white_offset = square_white_offset
        self.square_black_offset = square_black_offset
        self.class_ = class_
        self.id = id
        if start_stop is None:
            # render 2 octaves by default
            self.sns = SpecificNoteSet.from_noterange(SpecificNote('C', 0), SpecificNote('B', 1))
        else:
            # ensure that start and stop are white keys
            self.sns = SpecificNoteSet.from_noterange(
                start=start_stop[0] - 1 if start_stop[0].is_black else start_stop[0],
                stop=start_stop[1] + 1 if start_stop[1].abstract.is_black else start_stop[1],
            )
        self.white_notes = tuple(note for note in self.sns if not note.is_black)
        self.black_notes = tuple(note for note in self.sns if note.is_black)
        self.width = ww * len(self.white_notes)
        self.height = wh
        self.elements: list[svg.Element] = []
        self.notes = self.white_notes + self.black_notes
        self.make_piano()

    def make_piano(self) -> None:
        for note in self.notes:
            x, y, w, h, c, sx, sy = self.coord_helper(note)

            note_rect = svg.Rect(
                class_=['note', str(note)],
                x=x,
                y=y,
                width=w,
                height=h,
                fill=c.css_hex,
                stroke_width=1,
                stroke=config.BLACK_PALE.css_hex,
                onclick=self.note_onclicks.get(note, self.note_onclicks.get(note.abstract)),
            )
            # draw key

            if note_href := self.note_hrefs.get(note, self.note_hrefs.get(note.abstract)):
                self.elements.append(svg.A(href=note_href, elements=[note_rect]))
            else:
                self.elements.append(note_rect)

            # draw rectangle on top of note
            if rect_color := self.top_rect_colors.get(note, self.top_rect_colors.get(note.abstract)):
                self.elements.append(svg.Rect(class_=['top_rect', str(note)], x=x, y=y, width=w, height=self.top_rect_height, fill=rect_color.css_hex))

            # draw squares on notes
            if payload := self.squares.get(note, self.squares.get(note.abstract)):
                sq_elements: list[svg.Element] = []
                sq_rect = svg.Rect(
                    class_=['square', str(note)],
                    x=sx,
                    y=sy,
                    width=self.square_size,
                    height=self.square_size,
                    fill=payload.get('fill_color', config.WHITE_BRIGHT).css_hex,
                    stroke_width=1,
                    stroke=payload.get('border_color', config.BLACK_BRIGHT).css_hex,
                )
                sq_elements.append(sq_rect)

                if text := payload.get('text'):
                    sq_text = svg.Text(
                        class_=['square', str(note)],
                        x=sx + self.square_size // 2,
                        y=sy + self.square_size // 2,
                        font_family=payload.get('square_font_family', 'monospace'),  # type: ignore[arg-type]
                        font_size=payload.get('text_size', 15),
                        fill=payload.get('text_color', config.BLACK_BRIGHT).css_hex,
                        text=text,
                        text_anchor='middle',
                        dominant_baseline='central',
                    )
                    sq_elements.append(sq_text)
                self.elements.append(svg.G(class_=['square', str(note)], onclick=payload.get('onclick'), elements=sq_elements))

    def coord_helper(self, note: SpecificNote) -> tuple[int, int, int, int, Color, int, int]:
        """
        helper function which computes values for a given note

        Returns
        -------
        x: x coordinate of note rect
        y: y coordinate of note rect
        w: width of note rect
        h: height of note rect
        c: color of note
        sx: x coordinate of square
        sy: x coordinate of square
        """
        c = self.note_colors.get(note, self.note_colors.get(note.abstract, note_color(note)))
        x0 = 0
        y = 0
        if note in self.white_notes:
            x = x0 + self.ww * self.white_notes.index(note)
            sx = (x + x + self.ww) // 2 - self.square_size // 2
            sy = self.wh - self.square_size - self.square_white_offset
            return x, y, self.ww, self.wh, c, sx, sy
        if note in self.black_notes:
            x = x0 + self.ww * self.white_notes.index(note + 1) - self.bw // 2
            sx = self.ww * self.white_notes.index(note + 1) - self.square_size // 2
            sy = self.bh - self.square_size - self.square_black_offset
            return x, y, self.bw, self.bh, c, sx, sy
        raise ValueError(f'note {note} is not in {self.sns}')

    @property
    def svg(self) -> svg.SVG:
        return svg.SVG(width=self.width, height=self.height, elements=self.elements, class_=list(self.class_), id=self.id)

    def _repr_svg_(self) -> str:
        return str(self.svg)



from __future__ import annotations

import operator
import warnings
from typing import TYPE_CHECKING

import svg
from colortool import Color

from musiclib import config
from musiclib.midi.pitchbend import make_notes_pitchbends
from musiclib.noteset import SpecificNoteSet
from musiclib.svg.isomorphic.piano import IsoPiano
from musiclib.svg.isomorphic.text import FromIntervalDict

if TYPE_CHECKING:
    from svg._types import Number

    from musiclib.midi.parse import Midi
    from musiclib.midi.parse import MidiNote
    from musiclib.midi.parse import MidiPitch
    from musiclib.note import SpecificNote


class PianoRoll:
    def __init__(
        self,
        midi: Midi,
        pitchbend_semitones: int = 2,
        key_width: int = 20,
        key_height: int = 400,
        time_signature: tuple[int, int] = (4, 4),
        grid_denominator: int = 4,
        start_stop: tuple[SpecificNote, SpecificNote] | None = None,
        note_font_size: int = 10,
    ) -> None:
        if time_signature != (4, 4):
            raise NotImplementedError('only time_signature=(4, 4) is supported')
        self.time_signature = time_signature
        if grid_denominator < 1:
            raise ValueError('grid_denominator should be >= 1')
        self.grid_denominator = grid_denominator
        if start_stop is None:
            start_stop = (
                min(midi.notes, key=operator.attrgetter('note')).note,
                max(midi.notes, key=operator.attrgetter('note')).note,
            )
        self.sns = SpecificNoteSet.from_noterange(*start_stop)
        self.key_width = key_width
        self.key_height = key_height
        self.piano = IsoPiano(
            n_cols=len(self.sns),
            interval_colors=dict.fromkeys(self.sns.intervals, config.WHITE_PALE),
            interval_strokes=dict.fromkeys(self.sns.intervals, {'stroke': config.BLACK_PALE, 'stroke_width': 0.5}),
            interval_text=FromIntervalDict({i: str(n) for i, n in zip(self.sns.intervals, self.sns.notes_ascending, strict=True)}),
            radius=key_width // 2,
            radius1=key_height // 2,
        )
        self.midi = midi
        self.duration = max(self.midi.notes, key=operator.attrgetter('off')).off
        self.pitchbend_semitones = pitchbend_semitones
        self.note_font_size = note_font_size
        self.elements = self.piano.elements.copy()
        self.draw_bar_lines()
        self.add_notes_elements()
        self.add_pitchbend_elements()

    def add_notes_elements(self) -> None:
        for i, note in enumerate(self.midi.notes):
            if note.note not in self.sns:
                warnings.warn(f'note {note.note} is not in {self.sns}')
                continue
            x = self.key_width * self.sns.index(note.note)
            h = (note.off - note.on) / self.duration * self.key_height
            y = note.on / self.duration * self.key_height
            y_svg = self.key_height - y - h
            self.elements.append(
                svg.Rect(
                    id=f'note-{i}',
                    class_=['note', str(note.note)],
                    x=x,
                    y=y_svg,
                    width=self.key_width,
                    height=h,
                    fill=Color(0xFC7B7B).css_hex,
                    stroke_width=1,
                    stroke=config.BLACK_PALE.css_hex,
                    onclick=f"select_note('note-{i}')",
                ),
            )
            self.elements.append(
                svg.Text(
                    id=f'note-{i}-text',
                    class_=['note-text', str(note.note)],
                    x=x + self.key_width / 2,
                    y=y_svg + h / 2,
                    text=str(note.note),
                    font_family='monospace',
                    font_size=self.note_font_size,
                    text_anchor='middle',
                    dominant_baseline='middle',
                ),
            )

    def add_pitchbend_elements(self) -> None:
        if not self.midi.pitchbend:
            return

        PITCHBEND_MAX = 8192  # noqa: N806

        defs = svg.Defs(
            elements=[
                svg.Marker(
                    id='circle',
                    markerWidth=4,
                    markerHeight=4,
                    refX=2,
                    refY=2,
                    elements=[svg.Circle(cx=2, cy=2, r=2, stroke='none', fill=config.BLUE.css_hex)],
                ),
            ],
        )
        self.elements.append(defs)

        def note_pitch_x(note: MidiNote, p: MidiPitch) -> int:
            return int(self.key_width * self.sns.index(note.note) + self.key_width // 2 + self.pitchbend_semitones * self.key_width * p.pitch / PITCHBEND_MAX)

        for note, pitches in make_notes_pitchbends(self.midi).items():
            if note.note not in self.sns:
                warnings.warn(f'note {note.note} is not in {self.sns}')
                continue
            polyline_points_notes: list[Number] = []
            for pitch in pitches:
                x = note_pitch_x(note, pitch)
                y = pitch.time / self.duration * self.key_height
                y_svg = self.key_height - y
                polyline_points_notes += [x, y_svg]
            self.elements.append(
                svg.Polyline(
                    points=polyline_points_notes,
                    stroke=config.GREEN.css_hex,
                    fill='none',
                    stroke_width=2,
                    marker_start='circle',
                    marker_mid='circle',
                    marker_end='circle',
                ),
            )

        def pitch_x(p: MidiPitch) -> int:
            return int(self.piano.width / 2 * (p.pitch / PITCHBEND_MAX + 1))

        polyline_points: list[Number] = []
        for pitch in self.midi.pitchbend:
            x = pitch_x(pitch)
            y = pitch.time / self.duration * self.key_height
            y_svg = self.key_height - y
            polyline_points += [x, y_svg]
        self.elements.append(
            svg.Polyline(
                points=polyline_points,
                stroke=config.BLUE.css_hex,
                fill='none',
                stroke_width=2,
                marker_start='url(#circle)',
                marker_mid='url(#circle)',
                marker_end='url(#circle)',
            ),
        )

    def draw_bar_lines(self) -> None:
        ticks_per_bar = self.midi.ticks_per_beat * self.time_signature[0]
        for i in range(0, self.duration, ticks_per_bar // self.grid_denominator):
            stroke_width = 0.5 if i % self.midi.ticks_per_beat == 0 else 0.125
            y = self.key_height - i / self.duration * self.piano.height
            self.elements.append(
                svg.Line(
                    x1=0,
                    x2=self.piano.width,
                    y1=y,
                    y2=y,
                    stroke=Color(0x404040).css_hex,
                    stroke_width=stroke_width,
                ),
            )

    @property
    def svg(self) -> svg.SVG:
        return svg.SVG(width=self.piano.width, height=self.piano.height, elements=self.elements)

    def _repr_svg_(self) -> str:
        return str(self.svg)



import cmath
from typing import Any


def are_all_none(*args: Any) -> bool:
    return all(arg is None for arg in args)


def is_any_not_none(*args: Any) -> bool:
    return any(arg is not None for arg in args)


def are_mutually_exclusive(*args: Any, exactly_one: bool = False) -> bool:
    n_not_none = sum(1 for arg in args if arg is not None)
    if exactly_one:
        return n_not_none == 1
    return n_not_none <= 1


def increment_duplicates(a: list[int]) -> list[int]:
    if not a:
        return []
    res = [a[0]]
    for num in a[1:]:
        if num <= res[-1]:
            res.append(res[-1] + 1)
        else:
            res.append(num)
    return res


def intervals_rotations(intervals: frozenset[int]) -> tuple[frozenset[int], ...]:
    out = [sorted(intervals)]
    for _ in range(len(intervals) - 1):
        x = out[-1]
        out.append([x[i] - x[1] for i in range(1, len(intervals))] + [x[0] + 12 - x[1]])
    return tuple(map(frozenset, out))


def named_intervals_rotations(
    intervals: set[int] | frozenset[int],
    name_prefix: str,
) -> dict[str, frozenset[int]]:
    return {f'{name_prefix}_{i}': fs for i, fs in enumerate(intervals_rotations(frozenset(intervals)))}


def vertex(
    x: float,
    y: float,
    radius: float,
    n: int,
    i: int,
    phase: float = 0,
) -> tuple[float, float]:
    theta = phase + 2 * cmath.pi * i / n
    p = complex(y, x) + radius * cmath.exp(1j * theta)
    return p.imag, p.real


def line_intersection(
    x00: float,
    y00: float,
    x01: float,
    y01: float,
    x10: float,
    y10: float,
    x11: float,
    y11: float,
) -> tuple[float | None, float | None]:
    """
    computes coordinates of intersection of 2 lines.
    Each line is defined by 2 points:
    input arguments:
    x00 - x coordinate of 1st point of 1st line
    y00 - y coordinate of 1st point of 1st line
    x01 - x coordinate of 2nd point of 1st line
    y01 - y coordinate of 2nd point of 1st line
    x10 - x coordinate of 1st point of 2nd line
    y10 - y coordinate of 1st point of 2nd line
    x11 - x coordinate of 2nd point of 2nd line
    y11 - y coordinate of 2nd point of 2nd line

    output: x, y - coordinates of intersection of 2 lines
    """
    # Calculate the slope and y-intercept of the first line
    if x01 - x00 != 0:
        m1 = (y01 - y00) / (x01 - x00)
        b1 = y00 - m1 * x00
    else:
        m1 = None
        b1 = x00

    # Calculate the slope and y-intercept of the second line
    if x11 - x10 != 0:
        m2 = (y11 - y10) / (x11 - x10)
        b2 = y10 - m2 * x10
    else:
        m2 = None
        b2 = x10

    # Check if the lines are parallel
    if m1 == m2:
        return None, None  # The lines are parallel, so no intersection

    # If the first line is vertical
    if m1 is None:
        x = b1
        y = m2 * x + b2  # type: ignore[operator]
        return x, y

    # If the second line is vertical
    if m2 is None:
        x = b2
        y = m1 * x + b1
        return x, y

    # Calculate the intersection point
    x = (b2 - b1) / (m1 - m2)
    y = m1 * x + b1

    return x, y



from typing import Any

from musiclib import config
from musiclib.svg.isomorphic.hexagonal import Hexagonal
from musiclib.svg.isomorphic.squared import Squared


class ReprSVGMixin:
    def _repr_svg_(self, **kwargs: Any) -> str:
        method = getattr(self, config.repr_svg_config['method'])  # type: ignore[call-overload]
        kw = {**config.repr_svg_config['kwargs'], **kwargs}  # type: ignore[dict-item]
        plane_cls = kw.get('plane_cls')
        if plane_cls is not None:
            kw['plane_cls'] = {'Hexagonal': Hexagonal, 'Squared': Squared}[plane_cls]
        return str(method(**kw))



from __future__ import annotations

import functools
import itertools
from typing import TYPE_CHECKING
from typing import Any

from musiclib.progression import Progression

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Hashable

    from musiclib.note import SpecificNote
    from musiclib.noteset import SpecificNoteSet
    from musiclib.scale import Scale


def chord_pair_check_cache(f: Callable[..., bool]) -> Callable[..., bool]:
    cache: dict[Hashable, bool] = {}
    cache_info = {'hits': 0, 'misses': 0, 'currsize': 0}

    @functools.wraps(f)
    def is_check_passed(a: SpecificNoteSet, b: SpecificNoteSet, *args: Any) -> bool:
        a, b = Progression((a, b)).transposed_to_C0
        key = a, b, *args
        cached = cache.get(key)
        if cached is not None:
            cache_info['hits'] += 1
            return cached
        cache_info['misses'] += 1
        cache_info['currsize'] += 1
        computed = f(a, b, *args)
        cache[key] = computed
        return computed
    is_check_passed._cache = cache  # type: ignore[attr-defined]
    is_check_passed._cache_info = cache_info  # type: ignore[attr-defined]
    return is_check_passed


@chord_pair_check_cache
def is_parallel_interval(a: SpecificNoteSet, b: SpecificNoteSet, interval: int, /) -> bool:
    """
    parallel in same voices!
    if there'are eg fifth in 1st and fifth in 2nd chord but not from same voices
    - then it allowed (aint considered parallel) (test it)

    a1 - b1
    a0 - b0
    todo: what about fifths + octave (eg C5 G6 -> F5 C6)
    """
    voice_transitions = tuple(zip(a, b, strict=True))
    for (a0, b0), (a1, b1) in itertools.combinations(voice_transitions, 2):
        if abs(a0 - a1) % 12 == interval == abs(b0 - b1) % 12:
            return True
    return False


@chord_pair_check_cache
def is_hidden_parallel(a: SpecificNoteSet, b: SpecificNoteSet, interval: int, /) -> bool:
    """
    hidden/direct parallel/consecutive interval is when:
        1. outer voices (lower and higher) go in same direction (instead of oblique or contrary motion)
        2. they approach param:interval
    voice leading rules often forbid hidden fifths and octaves (param:interval = 7, 0) (explanation: 12 % 12 == 0 octave equal to unison)
    """
    a_low, a_high = a[0], a[-1]
    b_low, b_high = b[0], b[-1]

    is_same_direction = (a_low < b_low and a_high < b_high) or (a_low > b_low and a_high > b_high)
    return is_same_direction and (b_high - b_low) % 12 == interval


@chord_pair_check_cache
def is_voice_crossing(a: SpecificNoteSet, b: SpecificNoteSet, /) -> bool:
    n = len(b)
    for i in range(n):
        upper = i < n - 1 and b[i] > a[i + 1]
        lower = i > 0 and b[i] < a[i - 1]
        if upper or lower:
            return True
    return False


@chord_pair_check_cache
def is_large_leaps(a: SpecificNoteSet, b: SpecificNoteSet, interval: int, /) -> bool:
    return any(abs(an - bn) > interval for an, bn in zip(a, b, strict=True))


@chord_pair_check_cache
def is_make_major_scale_leading_tone_resolving_semitone_up(
    a: SpecificNoteSet,
    b: SpecificNoteSet,
    s: Scale,
    /,
) -> bool:
    if s.intervalset.names != frozenset({'major'}):
        raise ValueError('pass major scale')
    leading_tone = next(note for note in a.notes if note.abstract == s.notes_ascending[-1])
    tonic = next(note for note in b.notes if note.abstract == s.root)
    return tonic - leading_tone == 1


def is_large_spacing(c: SpecificNoteSet, max_interval: int = 12, /) -> bool:
    return any(b - a > max_interval for a, b in itertools.pairwise(c))


def is_small_spacing(c: SpecificNoteSet, min_interval: int = 3, /) -> bool:
    return any(b - a < min_interval for a, b in itertools.pairwise(c))


def find_paused_voices(a: SpecificNoteSet, b: SpecificNoteSet, n_notes: int) -> tuple[int, ...] | tuple[()]:
    if n_notes == 0 or len(a) == len(b) == 0:
        return ()
    if max(len(a), len(b)) > n_notes:
        raise ValueError('both chords should hava number of notes <= n_notes')
    less_notes, more_notes = sorted((a, b), key=len)
    used = set()

    def _key(note0: SpecificNote, note1: SpecificNote) -> int:
        return abs(note0 - note1)

    for note in less_notes:
        nearest = min(more_notes, key=functools.partial(_key, note1=note))
        used.add(nearest)
    paused = tuple(i for i, note in enumerate(more_notes) if note not in used)
    return paused  # noqa: RET504



from __future__ import annotations

import functools
from typing import TYPE_CHECKING
from typing import Any
from typing import TypeVar

if TYPE_CHECKING:
    from collections.abc import Iterator

    import svg

from musiclib import config
from musiclib.intervalset import IntervalSet
from musiclib.note import Note
from musiclib.noteset import NoteSet
from musiclib.svg.card import PlanePiano
from musiclib.svg.reprsvg import ReprSVGMixin
from musiclib.util.cache import Cached

Self = TypeVar('Self', bound='Scale')


class Scale(Cached, ReprSVGMixin):
    def __init__(self, root: Note, intervalset: IntervalSet) -> None:
        if not isinstance(root, Note):
            raise TypeError(f'expected Note, got {type(root)}')
        if not isinstance(intervalset, IntervalSet):
            raise TypeError(f'expected IntervalSet, got {type(intervalset)}')
        self.root = root
        self.intervalset = intervalset
        self.notes = frozenset({root + interval for interval in intervalset})
        self.noteset = NoteSet(self.notes)
        _notes_octave_fit = sorted(self.notes)
        _root_i = _notes_octave_fit.index(root)
        self.notes_ascending = _notes_octave_fit[_root_i:] + _notes_octave_fit[:_root_i]
        self.note_to_interval = dict(zip(self.notes_ascending, intervalset.intervals_ascending, strict=False))
        self.bits_chromatic_notes = tuple(int(Note(note) in self.notes) for note in config.chromatic_notes)
        self.note_i = {note: i for i, note in enumerate(self.notes_ascending)}
        self._key = self.root, self.intervalset

    @classmethod
    def from_name(cls: type[Self], root: str | Note, name: str) -> Self:
        if isinstance(root, str):
            root = Note(root)
        elif not isinstance(root, Note):
            raise TypeError(f'expected str | Note, got {type(root)}')
        return cls(root, IntervalSet(config.name_to_intervals[name]))

    @classmethod
    def from_notes(cls: type[Self], root: Note, notes: frozenset[Note]) -> Self:
        if not isinstance(root, Note):
            raise TypeError(f'expected Note, got {type(root)}')
        if root not in notes:
            raise ValueError('scale root must be in scale notes')
        return cls(root, IntervalSet(frozenset(note - root for note in notes)))

    @classmethod
    def from_str(cls: type[Self], string: str) -> Self:
        if string == '':
            raise ValueError('scale string must not be empty')
        if string[-2] != '/':
            raise ValueError('scale string must ends with scale root, example "CDEFGAB/C"')
        root = Note(string[-1])
        notes = frozenset(Note(note) for note in string[:-2])
        return cls.from_notes(root, notes)

    @functools.cached_property
    def note_scales(self) -> dict[str, dict[Note, str]]:
        _note_scales: dict[str, dict[Note, str]] = {}
        for name, kind in self.intervalset.name_kinds.items():
            scales = config.scale_order[kind]
            _scale_i = scales.index(name)
            scales = scales[_scale_i:] + scales[:_scale_i]
            _note_scales[kind] = {}
            for note, scale in zip(self.notes_ascending, scales, strict=True):
                _note_scales[kind][note] = scale
        return _note_scales

    def nths(self, ns: frozenset[int]) -> tuple[Scale, ...]:
        return tuple(
            Scale.from_notes(self.notes_ascending[i], frozenset(self.notes_ascending[(i + n) % len(self)] for n in ns))
            for i in range(len(self))
        )

    def transpose_to_note(self, note: Note) -> Scale:
        return Scale(note, self.intervalset)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Scale):
            raise TypeError
        return self._key == other._key

    def __hash__(self) -> int:
        return hash(self._key)

    def __len__(self) -> int:
        return len(self.notes)

    def __getitem__(self, item: int) -> Note:
        return self.notes_ascending[item]

    def __iter__(self) -> Iterator[Note]:
        return iter(self.notes_ascending)

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, Note):
            raise TypeError
        return item in self.notes

    @property
    def str_names(self) -> str:
        return f"{self.root} {' '.join(sorted(self.intervalset.names))}"

    def __str__(self) -> str:
        return f"{''.join(note.name for note in self.notes_ascending)}/{self.root}"

    def __repr__(self) -> str:
        return f'Scale({self.root!r}, {self.intervalset!r})'

    def __getnewargs__(self) -> tuple[Note, IntervalSet]:
        return (self.root, self.intervalset)

    def svg_piano(self, **kwargs: Any) -> svg.SVG:
        from musiclib.svg.card import Piano
        kwargs.setdefault('class_', tuple(self.intervalset.names))
        kwargs.setdefault('header_kwargs', {'title': f'{self.str_names}'})
        kwargs.setdefault(
            'regular_piano_kwargs', {
                'note_colors': {note: config.interval_colors[interval] for note, interval in self.note_to_interval.items()},
            },
        )
        return Piano(**kwargs).svg

    def svg_plane_piano(self, **kwargs: Any) -> svg.SVG:
        kwargs.setdefault('interval_colors', {i: config.interval_colors[i] for i in self.intervalset.intervals})
        kwargs.setdefault('header_kwargs', {'title': f'{self.str_names}'})
        return PlanePiano(**kwargs).svg



import asyncio
import functools

import mido
import mido.midifiles.midifiles

from musiclib.note import SpecificNote
from musiclib.noteset import SpecificNoteSet

Playable = SpecificNote | SpecificNoteSet


class Player:
    def __init__(self, midi_device: str | None = None) -> None:
        if midi_device is None:
            self.send_message = self._print_message
        else:
            self.port = mido.open_output(midi_device)
            self.send_message = self._send_message

    def _print_message(self, message: mido.Message) -> None:
        print('MIDI_DEVICE not found |', message)  # noqa: T201

    def _send_message(self, message: mido.Message) -> None:
        self.port.send(message)

    @functools.singledispatchmethod
    async def play(
        self,
        obj: Playable,  # pylint: disable=unused-argument
        seconds: float = 1,  # pylint: disable=unused-argument
        channel: int = 0,  # pylint: disable=unused-argument
        velocity: int = 100,  # pylint: disable=unused-argument
    ) -> None:
        ...

    @play.register
    async def _(
        self,
        obj: SpecificNote,
        seconds: float = 1,
        channel: int = 0,
        velocity: int = 100,
    ) -> None:
        self.send_message(mido.Message(type='note_on', channel=channel, note=obj.i, velocity=velocity))
        await asyncio.sleep(seconds)
        self.send_message(mido.Message(type='note_off', channel=channel, note=obj.i, velocity=velocity))

    @play.register
    async def _(
        self,
        obj: SpecificNoteSet,
        seconds: float = 1,
        channel: int = 0,
        velocity: int = 100,
    ) -> None:
        tasks = [self.play(note, seconds, channel, velocity) for note in obj.notes]
        await asyncio.gather(*tasks)

    def play_midi(self, midi: mido.MidiFile, beats_per_minute: float = 120) -> None:
        default_tempo = mido.midifiles.midifiles.DEFAULT_TEMPO
        mido.midifiles.midifiles.DEFAULT_TEMPO = mido.bpm2tempo(beats_per_minute)
        for message in midi.play():
            self.send_message(message)
        mido.midifiles.midifiles.DEFAULT_TEMPO = default_tempo

    async def aio_play_midi(self, midi: mido.MidiFile, beats_per_minute: float = 120) -> None:
        """For some reason this async method is playing with weird delays and time glitches unlike play_midi()"""
        default_tempo = mido.midifiles.midifiles.DEFAULT_TEMPO
        mido.midifiles.midifiles.DEFAULT_TEMPO = mido.bpm2tempo(beats_per_minute)
        for message in midi:
            self.send_message(message)
            await asyncio.sleep(message.time)
        mido.midifiles.midifiles.DEFAULT_TEMPO = default_tempo



import collections
import dataclasses
import itertools
from typing import no_type_check

import numpy as np

from musiclib.midi.parse import Midi
from musiclib.midi.parse import MidiNote
from musiclib.midi.parse import MidiPitch
from musiclib.midi.parse import index_abs_messages
from musiclib.tempo import Tempo
from musiclib.util.etc import increment_duplicates


@dataclasses.dataclass
class PitchPattern:
    time_bars: list[float]
    pitch_st: list[float]


def interpolate_pattern(pattern: PitchPattern, n_interp_points: int) -> PitchPattern:
    if n_interp_points <= len(pattern.time_bars):
        raise ValueError(f'n_interp_points should be > len(pattern.time_bars), got {n_interp_points=} and {len(pattern.time_bars)=}')
    original_points_list = list(zip(pattern.time_bars, pattern.pitch_st, strict=True))
    original_points_indices = {point: i for i, point in enumerate(original_points_list)}
    original_points = set(original_points_list)
    new_t = np.linspace(pattern.time_bars[0], pattern.time_bars[-1], num=n_interp_points).tolist()
    new_p = np.interp(new_t, pattern.time_bars, pattern.pitch_st).tolist()
    new_points = original_points.copy()
    for t, p in zip(new_t, new_p, strict=True):
        if len(new_points) >= n_interp_points:
            break
        if (t, p) in original_points:
            continue
        new_points.add((t, p))
    new_points_sorted = sorted(new_points, key=lambda point: (original_points_indices.get(point, -1), point))
    new_t, new_p = zip(*new_points_sorted, strict=True)
    return PitchPattern(time_bars=list(new_t), pitch_st=list(new_p))


def insert_pitch_pattern(
    midi: Midi,
    time_ticks: int,
    pattern: PitchPattern,
    *,
    pitchbend_semitones: int = 2,
    n_interp_points: int | None = None,
    increment_duplicates_: bool = True,
) -> Midi:
    if n_interp_points is not None:
        pattern = interpolate_pattern(pattern, n_interp_points)
    time_bars = Tempo(ticks=time_ticks, ticks_per_beat=midi.ticks_per_beat).bars
    pitchbend = [
        MidiPitch(
            time=int((time_bars + t) * midi.ticks_per_beat * 4),
            pitch=int(p / pitchbend_semitones * 8191),
        )
        for t, p in zip(pattern.time_bars, pattern.pitch_st, strict=True)
    ]
    pitchbend = sorted(midi.pitchbend + pitchbend, key=lambda p: p.time)
    if increment_duplicates_:
        pitchbend = [
            MidiPitch(time=time, pitch=pitch)
            for time, pitch in zip(
                increment_duplicates([p.time for p in pitchbend]),
                [p.pitch for p in pitchbend],
                strict=True,
            )
        ]
    return Midi(
        notes=midi.notes,
        pitchbend=pitchbend,
        ticks_per_beat=midi.ticks_per_beat,
    )


def make_notes_pitchbends(midi: Midi) -> dict[MidiNote, list[MidiPitch]]:
    T, P = zip(*[(e.time, e.pitch) for e in midi.pitchbend], strict=True)  # noqa: N806
    interp_t = []
    for note in midi.notes:
        interp_t += [note.on, note.off]
    interp_p = np.interp(interp_t, T, P, left=0).astype(int).tolist()  # https://docs.scipy.org/doc/scipy/tutorial/interpolate/1D.html#piecewise-linear-interpolation
    interp_pitches = sorted(midi.pitchbend + [MidiPitch(time=t, pitch=p) for t, p in zip(interp_t, interp_p, strict=True)], key=lambda p: p.time)
    midi_tmp = Midi(notes=midi.notes, pitchbend=interp_pitches)
    notes_pitchbends = collections.defaultdict(list)
    playing_notes = set()
    for im in index_abs_messages(midi_tmp):
        if im.message.type in {'note_on', 'note_off'}:
            note = midi.notes[im.index]
            if im.message.type == 'note_on':
                playing_notes.add(note)
            elif im.message.type == 'note_off':
                playing_notes.remove(note)
        elif im.message.type == 'pitchwheel':
            for note in playing_notes:
                notes_pitchbends[note].append(midi_tmp.pitchbend[im.index])
    return dict(notes_pitchbends)


@no_type_check
def add_pitchbend_from_overlapping_notes(midi: Midi, pitchbend_semitones: int = 2) -> Midi:
    notes_to_delete = set()
    pitchbend = []
    new_notes = []
    it = itertools.chain(midi.notes, [None])
    for a, b in itertools.pairwise(it):
        if b is None:
            new_notes.append(a)
            break
        if a in notes_to_delete:
            continue
        if a.off < b.on:
            new_notes.append(a)
            continue
        if b.note - a.note > pitchbend_semitones:
            raise ValueError(f'note leap {b.note - a.note} is larger than {pitchbend_semitones=}. Increase pitchbend_semitones')
        pitch = -int((b.note - a.note) / pitchbend_semitones * 8191)
        pitchbend.append(MidiPitch(time=a.on, pitch=pitch))
        pitchbend.append(MidiPitch(time=b.on, pitch=0))
        pitchbend.append(MidiPitch(time=b.off, pitch=0))
        new_notes.append(MidiNote(note=b.note, on=a.on, off=b.off))
        notes_to_delete.add(b)
    return Midi(notes=new_notes, pitchbend=pitchbend)



import svg
from svg._types import Number


class NestedSVG:
    def __init__(
        self,
        svgs: list[svg.SVG],
        coordinates: list[tuple[Number, Number]],
        width: Number = 600,
        height: Number = 400,
        class_: list[str] | None = None,
        id: str | None = None,  # noqa: A002 pylint: disable=redefined-builtin
    ) -> None:
        self.svgs = svgs
        self.coordinates = coordinates
        self.width = width
        self.height = height
        self.class_ = class_
        self.id = id

    @property
    def svg(self) -> svg.SVG:
        return svg.SVG(
            width=self.width,
            height=self.height,
            elements=[
                svg.SVG(
                    x=x,
                    y=y,
                    width=_svg.width,
                    height=_svg.height,
                    viewBox=svg.ViewBoxSpec(0, 0, _svg.width, _svg.height),  # type: ignore[arg-type]
                    elements=_svg.elements if isinstance(_svg, svg.SVG) else [_svg],
                    class_=_svg.class_,
                    id=_svg.id,
                )
                for _svg, (x, y) in zip(self.svgs, self.coordinates, strict=True)
            ],
            class_=self.class_,
            id=self.id,
        )

    def _repr_svg_(self) -> str:
        return str(self.svg)



import cmath
from typing import Literal

from musiclib.svg.isomorphic.base import IsomorphicKeyboard
from musiclib.util.etc import vertex


class Hexagonal(IsomorphicKeyboard):
    def add_keys(self) -> None:
        if self.rotated:
            for col in (self.col_range or range(-1, self.n_cols + 1)):
                for row in (self.row_range or range(-2, self.n_rows + 1, 2)):
                    self.add_key(row + col % 2, col)
            return
        for row in (self.row_range or range(-1, self.n_rows + 1)):
            for col in (self.col_range or range(-2, self.n_cols + 1, 2)):
                self.add_key(row, col + row % 2)

    @staticmethod
    def transform_coordinates(x: int, y: int) -> tuple[int, int]:
        return (x - y) // 2, y

    def row_col_to_interval(self, row: float, col: float) -> int:
        if self.rotated:
            ax0, ax1 = self.transform_coordinates(round(row), round(col))
        else:
            ax0, ax1 = self.transform_coordinates(round(col), round(row))
        return ax0 * self.ax0_step + ax1 * self.ax1_step

    def col_to_x(self, col: float) -> float:
        if self.rotated:
            return self._opposite_vertices_axis_index_to_px(col) + self.offset_x
        return self._opposite_midpoints_axis_index_to_px(col) + self.offset_x

    def row_to_y(self, row: float, *, invert_axis: bool = True) -> float:
        if invert_axis:
            return self.height - (self.row_to_y(row, invert_axis=False) + self.offset_y)
        if self.rotated:
            return self._opposite_midpoints_axis_index_to_px(row) + self.offset_y
        return self._opposite_vertices_axis_index_to_px(row) + self.offset_y

    @property
    def width(self) -> int:
        if self.rotated:
            return int(self.col_to_x(self.n_cols) - 0.5 * self.radius)
        return int(self.col_to_x(self.n_cols))

    @property
    def height(self) -> int:
        if self.rotated:
            return int(self.row_to_y(self.n_rows, invert_axis=False))
        return int(self.row_to_y(self.n_rows, invert_axis=False) - 0.5 * self.radius)

    def _opposite_vertices_axis_index_to_px(self, i: float) -> float:
        return self.radius * (i * 1.5 + 1)

    def _opposite_midpoints_axis_index_to_px(self, i: float) -> float:
        return self.h * (i + 1)

    @property
    def h(self) -> float:
        return 3 ** 0.5 / 2 * self.radius

    def key_points(self, x: float, y: float) -> list[float]:
        points: list[float] = []
        phase = cmath.pi / 6 if self.rotated else 0
        for i in range(7):
            points += vertex(x, y, self.radius, 6, i, phase)
        return points

    def ax_split_part_rect_coordinates(self, x: float, y: float, part: int, ax: Literal['horizontal', 'vertical']) -> dict[str, float]:
        ax_split_len = self.h if self.rotated else self.radius
        ax_other_len = self.radius if self.rotated else self.h
        z = ax_split_len * 2 / self.n_parts  # type: ignore[operator]
        if ax == 'vertical':
            return {
                'x': x - ax_other_len,
                'y': y + ax_split_len - (part + 1) * z,
                'width': ax_other_len * 2,
                'height': z,
            }
        if ax == 'horizontal':
            ax_split_len, ax_other_len = ax_other_len, ax_split_len
            z = ax_split_len * 2 / self.n_parts  # type: ignore[operator]
            return {
                'x': x - ax_split_len + part * z,
                'y': y - ax_other_len,
                'width': z,
                'height': ax_other_len * 2,
            }
        raise ValueError(f'Unsupported ax: {ax}, must be horizontal or vertical')



from typing import Any
from typing import Literal

from musiclib.svg.isomorphic.base import IsomorphicKeyboard


class IsoPiano(IsomorphicKeyboard):
    def __init__(
        self,
        radius1: int = 36,
        **kw: Any,
    ) -> None:
        self.radius1 = radius1
        if kw.get('rotated', False):
            kw.setdefault('n_cols', 1)
            kw.setdefault('n_rows', 12)
        else:
            kw.setdefault('n_rows', 1)
            kw.setdefault('n_cols', 12)
        super().__init__(**kw)

    def validate_dimensions(
        self,
        n_rows: int | None,
        n_cols: int | None,
        row_range: range | None,
        col_range: range | None,
    ) -> None:
        if self.rotated and n_cols != 1:
            raise ValueError('n_cols must be 1 for rotated IsoPiano')
        if not self.rotated and n_rows != 1:
            raise ValueError('n_rows must be 1 for IsoPiano')
        super().validate_dimensions(n_rows, n_cols, row_range, col_range)

    def add_keys(self) -> None:
        for row in (self.row_range or range(self.n_rows)):
            for col in (self.col_range or range(self.n_cols)):
                self.add_key(row, col)

    def row_col_to_interval(self, row: float, col: float) -> int:
        if self.rotated:
            return round(row)
        return round(col)

    def col_to_x(self, col: float) -> float:
        if self.rotated:
            return self.radius1 * (col * 2 + 1) + self.offset_x
        return self.radius * (col * 2 + 1) + self.offset_x

    def row_to_y(self, row: float) -> float:
        if self.rotated:
            return self.radius * (row * 2 + 1) + self.offset_y
        return self.radius1 * (row * 2 + 1) + self.offset_y

    @property
    def width(self) -> int:
        return int(self.col_to_x(self.n_cols - 0.5))

    @property
    def height(self) -> int:
        return int(self.row_to_y(self.n_rows - 0.5))

    def key_part_points(self, x: float, y: float, part: int) -> list[float]:
        raise NotImplementedError('TODO: split vertically')

    @staticmethod
    def vertex(x: float, y: float, radius_w: float, i: int, radius_h: float) -> tuple[float, float]:
        return {
            0: (x - radius_w, y - radius_h),
            1: (x + radius_w, y - radius_h),
            2: (x + radius_w, y + radius_h),
            3: (x - radius_w, y + radius_h),
        }[i % 4]

    def key_points(self, x: float, y: float) -> list[float]:
        points: list[float] = []
        for i in range(5):
            if self.rotated:
                points += self.vertex(x, y, self.radius1, i, self.radius)
            else:
                points += self.vertex(x, y, self.radius, i, self.radius1)
        return points

    def ax_split_part_rect_coordinates(self, x: float, y: float, part: int, ax: Literal['horizontal', 'vertical']) -> dict[str, float]:
        ax_split_len = self.radius if self.rotated else self.radius1
        ax_other_len = self.radius1 if self.rotated else self.radius
        z = ax_split_len * 2 / self.n_parts  # type: ignore[operator]
        if ax == 'vertical':
            return {
                'x': x - ax_other_len,
                'y': y + ax_split_len - (part + 1) * z,
                'width': ax_other_len * 2,
                'height': z,
            }
        if ax == 'horizontal':
            ax_split_len, ax_other_len = ax_other_len, ax_split_len
            z = ax_split_len * 2 / self.n_parts  # type: ignore[operator]
            return {
                'x': x - ax_split_len + part * z,
                'y': y - ax_other_len,
                'width': z,
                'height': ax_other_len * 2,
            }
        raise ValueError(f'Unsupported ax: {ax}, must be horizontal or vertical')



import collections
import functools
import operator
import typing as tp

from colortool import Color

from musiclib.interval import AbstractInterval
from musiclib.util.etc import named_intervals_rotations

NIDict: tp.TypeAlias = dict[str, frozenset[AbstractInterval]]

chromatic_notes = 'CdDeEFfGaAbB'
circle_of_fifths_clockwise = 'CGDAEBfdaebF'


name_to_intervals_kind_grouped = {
    'natural': {
        'major': {0, 2, 4, 5, 7, 9, 11},
        'dorian': {0, 2, 3, 5, 7, 9, 10},
        'phrygian': {0, 1, 3, 5, 7, 8, 10},
        'lydian': {0, 2, 4, 6, 7, 9, 11},
        'mixolydian': {0, 2, 4, 5, 7, 9, 10},
        'minor': {0, 2, 3, 5, 7, 8, 10},
        'locrian': {0, 1, 3, 5, 6, 8, 10},
    },
    'pentatonic': {
        'p_major': {0, 2, 4, 7, 9},
        'p_dorian': {0, 2, 5, 7, 10},
        'p_phrygian': {0, 3, 5, 8, 10},
        'p_mixolydian': {0, 2, 5, 7, 9},
        'p_minor': {0, 3, 5, 7, 10},
    },
    'messiaen': {
        'messiaen_0': {0, 2, 4, 6, 8, 10},
        'messiaen_1': {0, 2, 3, 5, 6, 8, 9, 11},
        'messiaen_2': {0, 1, 2, 4, 5, 6, 8, 9, 10},
        'messiaen_3': {0, 1, 2, 3, 6, 7, 8, 9},
        'messiaen_4': {0, 1, 2, 6, 7, 8},
        'messiaen_5': {0, 1, 2, 4, 6, 7, 8, 10},
        'messiaen_6': {0, 1, 2, 3, 4, 6, 7, 8, 9, 10},
    },
    'h_minor': named_intervals_rotations({0, 2, 3, 5, 7, 8, 11}, 'h_minor'),
    'h_major': named_intervals_rotations({0, 2, 4, 5, 7, 8, 11}, 'h_major'),
    'm_minor': named_intervals_rotations({0, 2, 3, 5, 7, 9, 11}, 'm_minor'),
    # chords: all have inversion number suffix to distinguish from scales
    # triads
    'major': named_intervals_rotations({0, 4, 7}, 'major'),
    'minor': named_intervals_rotations({0, 3, 7}, 'minor'),
    'dim': named_intervals_rotations({0, 3, 6}, 'dim'),
    # 7th
    'maj7': named_intervals_rotations({0, 4, 7, 11}, 'maj7'),
    '7': named_intervals_rotations({0, 4, 7, 10}, '7'),
    'min7': named_intervals_rotations({0, 3, 7, 10}, 'min7'),
    'half-dim7': named_intervals_rotations({0, 3, 6, 10}, 'half-dim7'),
    'dim7': named_intervals_rotations({0, 3, 6, 9}, 'dim7'),
    # 6th
    '6': named_intervals_rotations({0, 4, 7, 9}, '6'),
    'm6': named_intervals_rotations({0, 3, 7, 9}, 'm6'),
    # other
    'aug': named_intervals_rotations({0, 4, 8}, 'aug'),
    'sus2': named_intervals_rotations({0, 2, 7}, 'sus2'),
    'sus4': named_intervals_rotations({0, 5, 7}, 'sus4'),
}
name_to_intervals_kind_grouped: dict[str, NIDict] = {k: {kk: frozenset(map(AbstractInterval, v)) for kk, v in kv.items()} for k, kv in name_to_intervals_kind_grouped.items()}  # type: ignore[no-redef, attr-defined]
name_to_intervals: NIDict = functools.reduce(operator.or_, name_to_intervals_kind_grouped.values())  # type: ignore[arg-type]
_intervals_to_names = collections.defaultdict(set)
for n, i in name_to_intervals.items():
    _intervals_to_names[i].add(n)
intervals_to_names = {k: frozenset(v) for k, v in _intervals_to_names.items()}
name_to_intervals_key = {kind: frozenset(kv.values()) for kind, kv in name_to_intervals_kind_grouped.items()}  # type: ignore[attr-defined]
intervals_key_to_name = {v: k for k, v in name_to_intervals_key.items()}

scale_order = {}
kinds = {}
for kind, kv in name_to_intervals_kind_grouped.items():
    scale_order[kind] = tuple(kv.keys())  # type: ignore[attr-defined]
    kinds.update(dict.fromkeys(kv.keys(), kind))  # type: ignore[attr-defined]

nths = {
    'triads': frozenset({0, 2, 4}),
    'sixths': frozenset({0, 2, 4, 5}),
    'sevenths': frozenset({0, 2, 4, 6}),
    'ninths': frozenset({0, 2, 4, 6, 8}),
}

# colors
WHITE_PALE = Color(0xAAAAAA)
BLACK_PALE = Color(0x505050)
WHITE_BRIGHT = Color(0xFFFFFF)
BLACK_BRIGHT = Color(0x000000)
RED = Color(0xFF0000)
GREEN = Color(0x00FF00)
BLUE = Color(0x4F88EA)


interval_colors = {
    AbstractInterval(0): Color(0xFF0000),
    AbstractInterval(1): Color(0x800000),
    AbstractInterval(2): Color(0xff60a0),
    AbstractInterval(3): Color(0xfffd37),
    AbstractInterval(4): Color(0x00ff59),
    AbstractInterval(5): Color(0x00d5ff),
    AbstractInterval(6): Color(0x8FBC8F),
    AbstractInterval(7): Color(0x0000FF),
    AbstractInterval(8): Color(0xff7000),
    AbstractInterval(9): Color(0x007000),
    AbstractInterval(10): Color(0x9c5c02),
    AbstractInterval(11): Color(0xa000ff),
}

repr_svg_piano_config = {
    'method': 'svg_piano',
    'kwargs': {},
}
repr_svg_plane_config = {
    'method': 'svg_plane_piano',
    'kwargs': {
        'piano_kwargs': {},
    },
}

repr_svg_plane_config_complex_example = {
    'method': 'svg_plane_piano',
    'kwargs': {
        'n_rows': 10,
        'n_cols': 24,
        'plane_cls': 'Squared',
        'plane_kwargs': {
            'ax0_step': 1,
            'ax1_step': 7,
            'rotated': True,
        },
    },
}

repr_svg_config = repr_svg_plane_config



import typing as tp

import numpy as np

from musiclib.interval import AbstractInterval

SVG_TEXT_KW = dict[str, tp.Any]
TEXT_CALLABLE: tp.TypeAlias = tp.Callable[[int, int, float, float], SVG_TEXT_KW | None]
TEXT_DEFAULT_KW = {
    'font_family': 'monospace',
    'text_anchor': 'middle',
    'dominant_baseline': 'middle',
    'pointer_events': 'none',  # probably not needed when using event.preventDefault() on transparent polygon
}

IntervalDict: tp.TypeAlias = dict[AbstractInterval | int, str]


def middle_text_kw_abstract_interval(interval: int, radius: int, x: float, y: float) -> SVG_TEXT_KW:
    return {**TEXT_DEFAULT_KW, 'font_size': int(0.5 * radius), 'text': str(AbstractInterval(interval)), 'x': x, 'y': y}


def sub_text_kw_interval(interval: int, radius: int, x: float, y: float) -> SVG_TEXT_KW:
    return {**TEXT_DEFAULT_KW, 'font_size': int(0.3 * radius), 'text': np.base_repr(interval, base=12), 'x': x, 'y': y + int(0.4 * radius)}


class FromIntervalDict:
    def __init__(self, interval_dict: IntervalDict, *, abstract: bool = False) -> None:
        self.interval_dict = interval_dict
        self.abstract = abstract

    def __call__(self, interval: int, radius: int, x: float, y: float) -> SVG_TEXT_KW:

        return {
            **TEXT_DEFAULT_KW,
            'font_size': int(0.5 * radius),
            'text': self.interval_dict.get(AbstractInterval(interval) if self.abstract else interval),
            'x': x,
            'y': y,
        }



from __future__ import annotations

import functools
import itertools
import random
import statistics

from opseq import OpSeq

from musiclib.util.cache import Cached


class Rhythm(Cached):
    def __init__(
        self,
        notes: tuple[int, ...],
        bar_notes: int = 16,  # grid size
    ) -> None:
        self.notes = notes
        self.bar_notes = bar_notes
        self.bits = ''.join(map(str, self.notes))

    @classmethod
    def random_rhythm(cls, n_notes: int | None = None, bar_notes: int = 16) -> Rhythm:
        if n_notes is None:
            n_notes = random.randint(1, bar_notes)
        if not 0 < n_notes <= bar_notes:
            raise ValueError(f'number of notes should be more than 1 and less than bar_notes={bar_notes}')
        notes = [1] * n_notes + [0] * (bar_notes - n_notes)
        random.shuffle(notes)
        return cls(tuple(notes), bar_notes=bar_notes)

    def __repr__(self) -> str:
        return f"Rhythm('{self.bits}')"

    @functools.cached_property
    def has_contiguous_ones(self) -> bool:
        return (
            self.notes[0] == 1 and self.notes[-1] == 1
            or any(len(list(g)) > 1 for k, g in itertools.groupby(self.notes, key=bool) if k)
        )

    @staticmethod
    def have_no_contiguous_ones(prev: int, curr: int) -> bool:
        return not prev == curr == 1

    @functools.cached_property
    def score(self) -> float:
        """spacings variance"""
        first_1 = self.notes.index(1)
        x = self.notes[first_1:] + self.notes[:first_1]
        spacings = [len(list(g)) for k, g in itertools.groupby(x, key=bool) if not k]
        if len(spacings) == 1:  # TODO: try normalize into 0..1
            return float('inf')
        return statistics.variance(spacings)

    @staticmethod
    def all_rhythms(*, n_notes: int, bar_notes: int = 16, sort_by_score: bool = False) -> tuple[Rhythm, ...]:
        r0 = OpSeq(
            n=bar_notes,
            options=(0, 1),
            curr_prev_constraint={-1: Rhythm.have_no_contiguous_ones},
        )
        r1 = (r for r in r0 if sum(r) == n_notes)
        r2 = (Rhythm(r, bar_notes=bar_notes) for r in r1)
        if sort_by_score:
            r2 = sorted(((r.score, r) for r in r2), key=lambda sr: (sr[0], sr[1].notes))  # type: ignore[assignment]
        return tuple(r2)



import collections
import dataclasses
import functools
from collections.abc import Generator
from collections.abc import Iterable
from typing import Literal

import mido

from musiclib.note import SpecificNote
from musiclib.noteset import SpecificNoteSet
from musiclib.rhythm import Rhythm
from musiclib.tempo import Tempo


@dataclasses.dataclass(frozen=True)
@functools.total_ordering
class MidiNote:
    note: SpecificNote
    on: int
    off: int

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MidiNote):
            raise TypeError
        return self.on == other.on

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, MidiNote):
            raise TypeError
        return self.on < other.on

    def __len__(self) -> int:
        return self.off - self.on

    def __hash__(self) -> int:
        return hash((self.note, self.on, self.off))


@dataclasses.dataclass(frozen=True)
class MidiPitch:
    time: int
    pitch: int


@dataclasses.dataclass
class Midi:
    notes: list[MidiNote] = dataclasses.field(default_factory=list)
    pitchbend: list[MidiPitch] = dataclasses.field(default_factory=list)
    ticks_per_beat: int = 96

    def __len__(self) -> int:
        return max(note.off for note in self.notes)


@dataclasses.dataclass
class IndexedMessage:
    message: mido.Message
    index: int


def is_note(type_: Literal['on', 'off'], message: mido.Message) -> bool:
    """https://stackoverflow.com/a/43322203/4204843"""
    if type_ == 'on':
        return message.type == 'note_on' and message.velocity > 0  # type: ignore[no-any-return]
    if type_ == 'off':
        return message.type == 'note_off' or (message.type == 'note_on' and message.velocity == 0)  # type: ignore[no-any-return]
    raise ValueError


def parse_midi(midi: mido.MidiFile) -> Midi:
    if midi.type != 0:
        raise ValueError('only type 0 midi files are supported')
    track, = midi.tracks
    notes: list[MidiNote] = []
    pitchbend = []
    playing_notes = collections.defaultdict(dict)  # type: ignore[var-annotated]
    t = 0
    for message in track:
        t += message.time
        if is_note('on', message):
            playing_notes[message.note].update({'note': SpecificNote.from_i(message.note), 'on': t})
        elif is_note('off', message):
            note = playing_notes[message.note]
            note['off'] = t
            notes.append(MidiNote(**note))
            del playing_notes[message.note]
        elif message.type == 'pitchwheel':
            pitchbend.append(MidiPitch(time=t, pitch=message.pitch))
    return Midi(notes=notes, pitchbend=pitchbend, ticks_per_beat=midi.ticks_per_beat)


def midiobj_to_midifile(midi: Midi) -> mido.MidiFile:
    abs_messages = index_abs_messages(midi)
    t = 0
    messages = []
    for im in abs_messages:
        m = im.message.copy()
        m.time = im.message.time - t
        messages.append(m)
        t = im.message.time
    track = mido.MidiTrack(messages)
    return mido.MidiFile(type=0, tracks=[track], ticks_per_beat=midi.ticks_per_beat)


def index_abs_messages(midi: Midi) -> list[IndexedMessage]:
    """this are messages with absolute time, note real midi messages"""
    abs_messages = []
    for i, note in enumerate(midi.notes):
        abs_messages.append(IndexedMessage(message=mido.Message(type='note_on', time=note.on, note=note.note.i, velocity=100), index=i))
        abs_messages.append(IndexedMessage(message=mido.Message(type='note_off', time=note.off, note=note.note.i, velocity=100), index=i))
    for i, pitch in enumerate(midi.pitchbend):
        abs_messages.append(IndexedMessage(message=mido.Message(type='pitchwheel', time=pitch.time, pitch=pitch.pitch), index=i))
    # Sort by time. If time is equal sort using type priority in following order: note_on, pitchwheel, note_off
    abs_messages.sort(key=lambda m: (m.message.time, {'note_on': 0, 'pitchwheel': 1, 'note_off': 2}[m.message.type]))
    return abs_messages


def specific_note_set_to_midi(
    noteset: SpecificNoteSet,
    path: str | None = None,
    n_bars: int = 1,
) -> mido.MidiFile:
    mid = mido.MidiFile(type=0, ticks_per_beat=96)
    track = mido.MidiTrack()
    stop_time = int(n_bars * mid.ticks_per_beat * 4)
    for note in noteset.notes_ascending:
        track.append(mido.Message('note_on', note=note.i, velocity=100, time=0))
    for i, note in enumerate(noteset.notes_ascending):
        track.append(mido.Message('note_off', note=note.i, velocity=100, time=stop_time if i == 0 else 0))
    mid.tracks.append(track)
    mid.meta = {'noteset': noteset}
    if path is not None:
        mid.save(path)
    return mid


TO_MIDI_MUTUAL_EXCLUSIVE_ERROR = TypeError('note_, chord are mutually exclusive. Only 1 must be not None')


def rhythm_to_midi(  # noqa: C901
    rhythm: Rhythm,
    path: str | None = None,
    note_: SpecificNote | None = None,
    noteset: SpecificNoteSet | None = None,
    progression: Iterable[SpecificNoteSet] | None = None,
) -> mido.MidiFile:

    if note_ is not None and noteset is not None:
        raise TO_MIDI_MUTUAL_EXCLUSIVE_ERROR

    if noteset is None:
        if note_ is None:
            raise TO_MIDI_MUTUAL_EXCLUSIVE_ERROR
        note__ = note_

    if note_ is None and noteset is None:
        raise TO_MIDI_MUTUAL_EXCLUSIVE_ERROR

    tempo = Tempo()
    mid = mido.MidiFile(type=0, ticks_per_beat=tempo.ticks_per_beat)

    ticks_per_note = tempo.ticks_per_beat * tempo.beats_per_bar // rhythm.bar_notes
    track = mido.MidiTrack()
    t = 0

    def append_bar(noteset: SpecificNoteSet | None) -> None:
        nonlocal t
        for is_play in rhythm.notes:
            if is_play:
                notes = [note__.i] if noteset is None else [note.i for note in noteset.notes]
                for i, note in enumerate(notes):
                    track.append(mido.Message('note_on', note=note, velocity=100, time=t if i == 0 else 0))
                for i, note in enumerate(notes):
                    track.append(mido.Message('note_off', note=note, velocity=100, time=ticks_per_note if i == 0 else 0))
                t = 0
            else:
                t += ticks_per_note

    if progression is None:
        append_bar(noteset)
    else:
        for _noteset in progression:
            append_bar(_noteset)

    mid.tracks.append(track)
    if path is not None:
        mid.save(path)
    return mid


class SpecificNoteSetEvent:
    def __init__(
        self,
        sns: SpecificNoteSet,
        on: int,
        off: int,
    ) -> None:
        self.sns = sns
        self.on = on
        self.off = off

    def __len__(self) -> int:
        return self.off - self.on

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(sns={self.sns}, on={self.on}, off={self.off})'


def unique_notesets(midi: mido.MidiFile) -> Generator[SpecificNoteSetEvent, None, None]:
    if midi.type != 0:
        raise ValueError('only type 0 midi files are supported')
    track, = midi.tracks
    playing_notes = collections.defaultdict(dict)  # type: ignore[var-annotated]
    t = 0
    t_sns = 0
    for message in track:
        t += message.time
        if message.type not in {'note_on', 'note_off'}:
            continue
        sns = SpecificNoteSet(frozenset(n['note'] for n in playing_notes.values()))
        yield SpecificNoteSetEvent(sns=sns, on=t_sns, off=t)
        t_sns = t
        if is_note('on', message):
            playing_notes[message.note].update({'note': SpecificNote.from_i(message.note), 'on': t})
        elif is_note('off', message):
            note = playing_notes[message.note]
            note['off'] = t
            del playing_notes[message.note]


def to_dict(midi: mido.MidiFile) -> dict:  # type: ignore[type-arg]
    return {
        'type': midi.type,
        'ticks_per_beat': midi.ticks_per_beat,
        'tracks': [[message.dict() | {'is_meta': message.is_meta} for message in track]for track in midi.tracks],
    }


def from_dict(midi: dict) -> mido.MidiFile:  # type: ignore[type-arg]
    return mido.MidiFile(
        type=midi['type'],
        ticks_per_beat=midi['ticks_per_beat'],
        tracks=[
            mido.MidiTrack(
                (mido.MetaMessage if message.pop('is_meta') else mido.Message).from_dict(message)
                for message in track
            )
            for track in midi['tracks']
        ],
    )



from collections.abc import Iterable
from typing import Any

import svg
from colortool import Color

from musiclib import config
from musiclib.interval import AbstractInterval
from musiclib.svg.header import Header
from musiclib.svg.isomorphic.base import TEXT_CALLABLE
from musiclib.svg.isomorphic.base import middle_text_kw_abstract_interval
from musiclib.svg.isomorphic.hexagonal import Hexagonal
from musiclib.svg.isomorphic.piano import IsoPiano
from musiclib.svg.isomorphic.squared import Squared
from musiclib.svg.nested import NestedSVG
from musiclib.svg.piano import RegularPiano


class PlanePiano:
    def __init__(  # noqa: PLR0915
        self,
        *,
        interval_colors: dict[AbstractInterval | int, Color] | None = None,
        interval_strokes: dict[AbstractInterval | int, Color] | None = None,
        interval_radial_parts_colors: dict[int, dict[int, Color]] | None = None,
        interval_horizontal_parts_colors: dict[int, dict[int, Color]] | None = None,
        interval_vertical_parts_colors: dict[int, dict[int, Color]] | None = None,
        n_parts: int | None = None,
        interval_text: TEXT_CALLABLE | None = middle_text_kw_abstract_interval,
        interval_subtext: TEXT_CALLABLE | None = None,
        interval_extra_texts: Iterable[TEXT_CALLABLE] = (),
        n_rows: int = 4,
        n_cols: int = 24,
        radius: int = 18,
        width: int | None = None,
        height: int | None = None,
        class_: list[str] | None = None,
        id: str | None = None,  # noqa: A002 pylint: disable=redefined-builtin
        header_kwargs: dict[str, Any] | None = None,
        plane_kwargs: dict[str, Any] | None = None,
        plane_cls: type[Hexagonal] | type[Squared] = Hexagonal,
        piano_kwargs: dict[str, Any] | None = None,
    ) -> None:

        nested_svg_kw = {
            'svgs': [],
            'coordinates': [],
            'height': 0,
            'class_': class_,
            'id': id,
        }

        plane_kwargs = plane_kwargs.copy() if plane_kwargs is not None else {}
        plane_kwargs.setdefault('interval_colors', interval_colors)
        plane_kwargs.setdefault('interval_strokes', interval_strokes)
        plane_kwargs.setdefault('interval_radial_parts_colors', interval_radial_parts_colors)
        plane_kwargs.setdefault('interval_horizontal_parts_colors', interval_horizontal_parts_colors)
        plane_kwargs.setdefault('interval_vertical_parts_colors', interval_vertical_parts_colors)
        plane_kwargs.setdefault('n_parts', n_parts)
        plane_kwargs.setdefault('interval_text', interval_text)
        plane_kwargs.setdefault('interval_subtext', interval_subtext)
        plane_kwargs.setdefault('interval_extra_texts', interval_extra_texts)
        plane_kwargs.setdefault('radius', radius)
        plane_kwargs.setdefault('ax0_step', 2)
        plane_kwargs.setdefault('ax1_step', 1)
        plane_kwargs.setdefault('n_rows', n_rows)
        plane_kwargs.setdefault('n_cols', n_cols)
        self.plane = plane_cls(**plane_kwargs)

        if header_kwargs is not None:
            header_kwargs = header_kwargs.copy()
            header_kwargs.setdefault('width', self.plane.svg.width)
            self.header = Header(**header_kwargs)
            nested_svg_kw['svgs'].append(self.header.svg)  # type: ignore[attr-defined]
            nested_svg_kw['coordinates'].append((0, nested_svg_kw['height']))  # type: ignore[attr-defined]
            nested_svg_kw['height'] += self.header.svg.height  # type: ignore[operator]

        nested_svg_kw['svgs'].append(self.plane.svg)  # type: ignore[attr-defined]
        nested_svg_kw['coordinates'].append((0, nested_svg_kw['height']))  # type: ignore[attr-defined]
        nested_svg_kw['height'] += self.plane.svg.height  # type: ignore[operator]

        if piano_kwargs is not None:
            piano_kwargs = piano_kwargs.copy()
            piano_kwargs.setdefault('interval_colors', interval_colors)
            piano_kwargs.setdefault('interval_strokes', interval_strokes)
            # piano_kwargs.setdefault('interval_radial_parts_colors', interval_radial_parts_colors)
            # piano_kwargs.setdefault('interval_horizontal_parts_colors', interval_horizontal_parts_colors)
            # piano_kwargs.setdefault('interval_vertical_parts_colors', interval_vertical_parts_colors)
            # piano_kwargs.setdefault('n_parts', n_parts)
            piano_kwargs.setdefault('interval_text', interval_text)
            piano_kwargs.setdefault('interval_subtext', interval_subtext)
            piano_kwargs.setdefault('interval_extra_texts', interval_extra_texts)
            piano_kwargs.setdefault('col_range', range(-1, n_cols + 1))
            piano_kwargs.setdefault('ax0_step', 1)
            piano_kwargs.setdefault('ax1_step', 0)
            piano_kwargs.setdefault('n_rows', 1)
            piano_kwargs.setdefault('n_cols', None)

            if isinstance(self.plane, Hexagonal):
                if self.plane.rotated:
                    piano_kwargs.setdefault('radius', self.plane.radius * 3 / 4)
                    piano_kwargs.setdefault('radius1', self.plane.radius * 3 / 4)
                    piano_kwargs.setdefault('offset_x', self.plane.radius / 4)
                else:
                    piano_kwargs.setdefault('radius', self.plane.h / 2)
                    piano_kwargs.setdefault('radius1', self.plane.h / 2)
                    piano_kwargs.setdefault('offset_x', self.plane.h / 2)
            elif isinstance(self.plane, Squared):
                if self.plane.rotated:
                    piano_kwargs.setdefault('radius', self.plane.radius / 2)
                    piano_kwargs.setdefault('radius1', self.plane.radius / 2)
                    piano_kwargs.setdefault('offset_x', self.plane.radius / 2)
                else:
                    piano_kwargs.setdefault('radius', self.plane.h)
                    piano_kwargs.setdefault('radius1', self.plane.h)
                    piano_kwargs.setdefault('col_range', range(n_cols))
            else:
                raise ValueError(f'Unsupported plane_cls: {plane_cls}, must be Hexagonal or Squared')

            self.piano = IsoPiano(**piano_kwargs)
            nested_svg_kw['svgs'].append(self.piano.svg)  # type: ignore[attr-defined]
            nested_svg_kw['coordinates'].append((0, nested_svg_kw['height']))  # type: ignore[attr-defined]
            nested_svg_kw['height'] += self.piano.svg.height  # type: ignore[operator]

        if height is not None:
            nested_svg_kw['height'] = height
        nested_svg_kw['width'] = width or self.plane.svg.width
        self.nested_svg = NestedSVG(**nested_svg_kw)  # type: ignore[arg-type]

    @property
    def svg(self) -> svg.SVG:
        return self.nested_svg.svg

    def __str__(self) -> str:
        return str(self.svg)

    def _repr_svg_(self) -> str:
        return str(self)


class Piano:
    def __init__(
        self,
        *,
        margin: tuple[int, int, int, int] = (3, 3, 3, 3),
        padding: tuple[int, int, int, int] = (2, 2, 2, 2),
        shadow_offset: int = 2,
        border_radius: int = 3,
        background_color: Color = config.WHITE_BRIGHT,
        header_kwargs: dict[str, Any] | None = None,
        regular_piano_kwargs: dict[str, Any] | None = None,
        class_: list[str] | None = None,
        id: str | None = None,  # noqa: A002 # pylint: disable=redefined-builtin
    ) -> None:
        header_kwargs = header_kwargs.copy() if header_kwargs is not None else {}
        header_kwargs.setdefault('header_rect', False)
        header_kwargs.setdefault('margin', (0, 0, 0, 0))
        header_kwargs.setdefault('height', 30)
        self.header = Header(**header_kwargs)
        self.piano = RegularPiano(**(regular_piano_kwargs or {}))
        card_width = self.piano.width + padding[1] + padding[3]
        card_height = self.header.svg.height + self.piano.height + padding[0] + padding[2]  # type: ignore[operator]
        width = margin[3] + card_width + shadow_offset + margin[1]
        height = margin[0] + card_height + shadow_offset + margin[2]
        self.shadow_rect = svg.SVG(
            elements=[
                svg.Rect(
                    class_=['shadow_rect'],
                    x=margin[3] + shadow_offset,
                    y=margin[0] + shadow_offset,
                    width=card_width,
                    height=card_height,
                    fill=config.BLACK_BRIGHT.css_hex,
                    rx=border_radius,
                    ry=border_radius,
                ),
            ],
            class_=['shadow_rect_svg'],
            width=width,
            height=height,
        )
        self.card_rect = svg.SVG(
            elements=[
                svg.Rect(
                    x=margin[3],
                    y=margin[0],
                    class_=['card_rect'],
                    width=card_width,
                    height=card_height,
                    fill=background_color.css_hex,
                    rx=border_radius,
                    ry=border_radius,
                    stroke_width=1,
                    stroke=config.BLACK_PALE.css_hex,
                ),
            ],
            class_=['card_rect_svg'],
            width=width,
            height=height,
        )
        self.nested_svg = NestedSVG(
            svgs=[
                self.shadow_rect,
                self.card_rect,
                self.header.svg,
                self.piano.svg,
            ],
            coordinates=[
                (0, 0),
                (0, 0),
                (margin[3] + padding[3], margin[0]),
                (margin[3] + padding[3], margin[0] + padding[0] + self.header.svg.height),  # type: ignore[operator]
            ],
            width=width,
            height=height,
            class_=class_,
            id=id,
        )

    @property
    def svg(self) -> svg.SVG:
        return self.nested_svg.svg

    def __str__(self) -> str:
        return str(self.svg)

    def _repr_svg_(self) -> str:
        return str(self)



import functools
from typing import TypeVar

import numpy as np

from musiclib.util.cache import Cached

Self = TypeVar('Self', bound='AbstractInterval')


@functools.total_ordering
class AbstractInterval(Cached):
    def __init__(self, interval: int) -> None:
        if not isinstance(interval, int):
            raise TypeError(f'expected int, got {type(interval)}')
        self.interval = interval % 12

    @classmethod
    def from_str(cls: type[Self], string: str) -> Self:
        if string == '':
            raise ValueError('interval string must not be empty')
        return cls(int(string, base=12))

    def __neg__(self: Self) -> Self:
        return self.__class__(-self.interval)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AbstractInterval):
            return False
        return self.interval == other.interval

    def __hash__(self) -> int:
        return hash(self.interval)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, AbstractInterval):
            raise TypeError
        return self.interval < other.interval

    def __getnewargs__(self) -> tuple[int]:
        return (self.interval,)

    def __str__(self) -> str:
        return np.base_repr(self.interval, base=12)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self})'
